#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "nexus.h"
#include "pnx.h"

typedef FILE *InputStream;

static int pnx_package_args_init(int *argc, char ***argv)
{
    return 0;
}

static void pnx_usage_message()
{
}

static int pnx_new_process_params(char *buf, int size)
{
    char tmp_buf1[1024];
    int n_added;
    SV *sv, *name_sv;
    char *script_name;
    STRLEN len;

    /*
     * Snag the name of the script
     */
    name_sv = perl_get_sv("0", FALSE);
    script_name = (char *)SvPV(name_sv, len);
/*    printf("scriptname is '%s' len=%d\n", script_name, len);*/

    n_added = len + 9;
    if (n_added > size)
    {
	nexus_fatal("new_process_params(): Internal error: Not enough room in buffer for arguments\n");
    }

    strcpy(buf, "-script ");
    strncat(buf, script_name, len);
    buf[len + 8] = ' ';
    buf[len + 9] = '\0';

    return (n_added);
}

MODULE = Nexus		PACKAGE = Nexus::MessageBuffer	PREFIX = pnx_mb_

void
pnx_mb_pack(bufref, ...)
	SV *	bufref
	CODE:
          {
	    croak("Nexus::MessageBuffer::pack() no longer allowed.");
/*              pnx_mb_pack(bufref, &(ST(1)), items - 1);*/
          }

MODULE = Nexus		PACKAGE = Nexus::StashedMessageBuffer	PREFIX = pnx_smb_

void
pnx_smb_unpack(bufref)
	SV *	bufref
	PPCODE:
	{
	    SV **items;
	    int n_items;
	    int i;

	    croak("Nexus::StashedMessageBuffer::unpack() no longer allowed.");
#if 0
	    pnx_smb_unpack(bufref, &items, &n_items);

	    EXTEND(sp, n_items);
	    for (i = 0; i < n_items; i++)
	    {
		PUSHs(items[i]);
	    }
	    NexusFree(items);
#endif
	}

MODULE = Nexus		PACKAGE = Nexus		PREFIX = nexus_

char *
nexus_database_lookup(node_name, key)
	char *	node_name
	char *	key
	CODE:
		RETVAL = nexus_database_lookup(node_name, 0, key);
	OUTPUT:
	RETVAL

double
nexus_wallclock()

int
nexus_isa_string(val)
	SV *	val
	CODE:
		RETVAL = SvPOKp(SvRV(val));
	OUTPUT:
	RETVAL

int
nexus_isa_int(val)
	SV *	val
	CODE:
		RETVAL = SvIOKp(SvRV(val));
	OUTPUT:
	RETVAL

int
nexus_isa_double(val)
	SV *	val
	CODE:
		RETVAL = SvNOKp(SvRV(val));
	OUTPUT:
	RETVAL

int
nexus_isa(val, clas)
	SV *	val
	char *	clas
	CODE:
		RETVAL = sv_isa(val, clas);
	OUTPUT:
	RETVAL

void
nexus__exit(rc, shutdown)
	int	rc
	int	shutdown
	CODE:
		nexus_exit(rc, shutdown);

MODULE = Nexus		PACKAGE = Nexus		PREFIX = pnx_

InputStream
pnx_get_notify_fh()


void
pnx_sv_dump(sv)
	SV *	sv

void
pnx__poll(block)
	int	block

void
pnx_register_fd_notifiers(add, add_arg, remove, remove_arg)
	SV *	add
	SV *	add_arg
	SV *	remove
	SV *	remove_arg

SV *
pnx_init_rsr(gpref, subroutine)
	SV *	gpref
	char *	subroutine

void
pnx_init()
	PPCODE:
	{
	    nexus_node_t *nodes;
	    int count;
	    SV *sv;
	    AV *av;
	    int i;
	    int n_argc;
	    char **n_argv, *arg;
	    SV *name_sv;
	    char *script_name;
	    STRLEN len;
	    
	    /*
	     * Snag the name of the script and construct an argument list
	     */
	    name_sv = perl_get_sv("0", FALSE);
	    script_name = (char *)SvPV(name_sv, len);
/*	    printf("init: scriptname is '%s' len=%d\n", script_name, len);*/
/*	    _nx_set_executable_name(script_name, len);*/

	    n_argc = 0;
	    n_argv = (char **) NULL;
	    
	    av = perl_get_av("main::ARGV", FALSE);
	    if (av)
	    {
		n_argc = av_len(av) + 2;
/*		printf("argc=%d\n", n_argc);*/
		if (n_argc > 0)
		{
		    n_argv = (char **) malloc(sizeof(char *) * (n_argc + 1));

		    n_argv[0] = (char *) malloc(len + 1);
		    strncpy(n_argv[0], script_name, len);
		    n_argv[0][len] = '\0';
		    for (i = 1; i < n_argc; i++)
		    {
			sv = *(av_fetch(av, i - 1, 0));
			arg = SvPV(sv, len);
			n_argv[i] = (char *) malloc(len + 1);
			strncpy(n_argv[i], arg, len);
			n_argv[i][len] = '\0';
		    }
		    n_argv[n_argc] = (char *) NULL;

		    /*
		    printf("args: argc=%d\n", n_argc);
		    for (i = 0; i < n_argc; i++)
		    {
			printf("%d: <%s>\n", i, n_argv[i]);
		    }
		    */
		}
	    }

	    nexus_init(&n_argc, &n_argv,
		       "NPERL_ARGS", "nx",
		       pnx_package_args_init, pnx_usage_message,
		       pnx_new_process_params, (nexus_module_list_t *) NULL,
		       &nodes, &count);

	    if (n_argv)
	    {
		/*
		printf("args after init: argc=%d\n", n_argc);
		for (i = 0; i < n_argc; i++)
		{
		    printf("%d: <%s>\n", i, n_argv[i]);
		}
		*/
		

		/*
		 * Now reconstruct @ARGV with the new argument list
		 */
		 
		 av = perl_get_av("main::ARGV", FALSE);
		 if (av)
		 {
		     av_clear(av);
		     for (i = 1; i < n_argc; i++)
		     {
			 av_store(av, i - 1, newSVpv(n_argv[i], 0));
		     }
		 }
	    }

	    for (i = 0; i < count; i++)
	    {
		/*
	        printf("Got gp %s\n", nexus_global_pointer_string(&nodes[i].gp));
		*/

		nexus_global_pointer_t *gp;
		
		NexusMalloc(pnx_init, gp, nexus_global_pointer_t *,
			    sizeof(nexus_global_pointer_t));

		*gp = nodes[i].gp;
		XPUSHs(sv_setref_iv(newSVsv(&sv_undef),
				    "Nexus::GlobalPointer",
				    (IV) gp));
	    }

	}

void
pnx_start()
	CODE:
	{
	    nexus_start();
	}

void
pnx_start_nonblocking()
	CODE:
	{
	    nexus_start_nonblocking();
	}

SV *
pnx_create_context(gpref, pathname, size)
	SV *	gpref
	char *	pathname
	int	size

void
pnx_destroy_current_context()
	CODE:
	{
	    nexus_destroy_current_context(0);
	}

int
pnx_master_node()
	CODE:
	RETVAL = nexus_master_node();
	OUTPUT:
	RETVAL

void
pnx_set_debug_level(level)
	int	level
	CODE:
          {
#ifdef NEXUS_DEBUG
             extern int _nx_debug_level;
	     _nx_debug_level = level;
#endif
          }

void
pnx_set_attachment(url, ref)
	char *	url
	SV *	ref

void
pnx_remove_attachment(url)
	char *	url

HV *
pnx_get_attachment_table()

void
pnx_set_default_attachment(ref)
	SV *	ref

void
pnx_remove_default_attachment()

SV *
pnx_allow_attach(port, func)
	int	port
	SV *	func
          PPCODE:
          {
	      int rc;
	      char *host;
	      unsigned short pport;

	      pport = (unsigned short) port;
	      rc = pnx_allow_attach(&pport, &host, func);
	      port = (int) pport;
	      EXTEND(sp, 3);
	      PUSHs(sv_2mortal(newSViv(rc)));
	      if (rc == 0)
	      {
		  PUSHs(sv_2mortal(newSViv(port)));
		  PUSHs(sv_2mortal(newSVpv(host, strlen(host))));
	      }
	      else
	      {
		  PUSHs(&sv_undef);
		  PUSHs(&sv_undef);
	      }
          }

void
pnx_disallow_attach(port)
	int	port

SV *
pnx_attach(url)
	char *	url

void
pnx_enable_fault_tolerance(callback_func)
	SV *	callback_func

MODULE = Nexus		PACKAGE = Nexus::GlobalPointer	PREFIX = pnx_gp_

SV *
pnx_gp_new(clas, var)
	char *	clas
	SV *	var

SV *
pnx_gp_dereference(self)
	SV *	self

void
pnx_gp_print(self)
	SV *	self

char *
pnx_gp_as_string(self, ...)
	SV *	self
	PPCODE:
	{
	    {
		SV *	self = ST(0);
		char *	RETVAL;

		RETVAL = pnx_gp_as_string(self);
		ST(0) = sv_newmortal();
		sv_setpv((SV*)ST(0), RETVAL);
	    }
	    XSRETURN(1);
	}

int
pnx_gp_is_local(self)
	SV *	self

int
pnx_gp_test(self)
	SV *	self

void
pnx_gp_DESTROY(self)
	SV *	self

MODULE = Nexus		PACKAGE = Nexus::MessageBuffer	PREFIX = pnx_mb_

void
pnx_mb_sizeof_int(self, n_ints)
	SV *	self
	int	n_ints
          PPCODE:
          {
	      int size, n_elts;
	      pnx_mb_sizeof_int(self, n_ints, &size, &n_elts);
	      EXTEND(sp, 2);
	      PUSHs(sv_2mortal(newSViv(size)));
	      PUSHs(sv_2mortal(newSViv(n_elts)));
          }

void
pnx_mb_sizeof_double(self, n_doubles)
	SV *	self
	int	n_doubles
          PPCODE:
          {
	      int size, n_elts;
	      pnx_mb_sizeof_double(self, n_doubles, &size, &n_elts);
	      EXTEND(sp, 2);
	      PUSHs(sv_2mortal(newSViv(size)));
	      PUSHs(sv_2mortal(newSViv(n_elts)));
          }

void
pnx_mb_sizeof_char(self, n_chars)
	SV *	self
	int	n_chars
          PPCODE:
          {
	      int size, n_elts;
	      pnx_mb_sizeof_char(self, n_chars, &size, &n_elts);
	      EXTEND(sp, 2);
	      PUSHs(sv_2mortal(newSViv(size)));
	      PUSHs(sv_2mortal(newSViv(n_elts)));
          }

void
pnx_mb_sizeof_global_pointer(self, gp)
	SV *	self
	SV *	gp
          PPCODE:
          {
	      int size, n_elts;
	      pnx_mb_sizeof_global_pointer(self, gp, &size, &n_elts);
	      EXTEND(sp, 2);
	      PUSHs(sv_2mortal(newSViv(size)));
	      PUSHs(sv_2mortal(newSViv(n_elts)));
          }

void
pnx_mb_set_size(self, size, n_elts)
	SV *	self
	int	size
	int	n_elts

void
pnx_mb_put_int(self, intval)
	SV *	self
	int	intval

void
pnx_mb_put_double(self, dblval)
	SV *	self
	double	dblval

void
pnx_mb_put_string(self, string)
	PPCODE:
    {
	SV *	self = ST(0);
	STRLEN len;	
	char *	string = (char *)SvPV(ST(1),len);

	pnx_mb_put_string(self, string, (int) len);
    }

void
pnx_mb_put_global_pointer(self, gp)
	SV *	self
	SV *	gp

int
pnx_mb_send_rsr(self)
	SV *	self

void
pnx_mb_DESTROY(self)
	SV *	self

MODULE = Nexus		PACKAGE = Nexus::StashedMessageBuffer	PREFIX = pnx_smb_

void
pnx_smb_get_int(self, intref)
	SV *	self
	SV *	intref

void
pnx_smb_get_double(self, doubleref)
	SV *	self
	SV *	doubleref

void
pnx_smb_get_string(self, stringref, len)
	SV *	self
	SV *	stringref
	int	len

void
pnx_smb_get_global_pointer(self, gpref)
	SV *	self
	SV *	gpref

void
pnx_smb_DESTROY(self)
	SV *	self

