package Nexus::DeepPacker;

use Carp;

$debug = 0;

#
# Packed data structure format
#
# packed_ds: <N items> [items]
#
# item: <type> <n_items> lowlevel_data
#
# type:
#	i	int
#	d	double
#	c	char
#	a	array
#	h	hash
#	r	rref
#
# Notes
#
# The n_chars++ in each of the pack routines is to allocate space for
# the type specifier.
#

sub pack
{
    my($buf, @args) = @_;
    my($arg, $packed_arg);
    my($pack_info, $i);
    my($size, $n_elts);

    $pack_info = new Nexus::DeepPacker::PackInfo($buf);

    &pack_array(\@args, $pack_info);

    &dump_packed($pack_info) if $debug;

    $size = $pack_info->{size};
    $n_elts = $pack_info->{n_elts};
    print "computed size=$size n_elts=$n_elts\n" if $debug;

    $buf->set_size($pack_info->{size}, $pack_info->{n_elts});
    &pack_buffer($buf, $pack_info);
    
    return $pack_info;
}

sub compute_size
{
    my($buf, $pack_info) = @_;
    my($size, $n_elts);
    my($s, $e, $gp);

    $size = $n_elts = 0;

    print "compute: buf=$buf\n" if $debug;
    ($s, $e) = $buf->sizeof_int($pack_info->{n_ints});
    $size += $s;
    $n_elts += $e;
    print "int size: $s $e\n";

    ($s, $e) = $buf->sizeof_char($pack_info->{n_chars});
    $size += $s;
    $n_elts += $e;
    print "char size: $s $e\n";

    ($s, $e) = $buf->sizeof_double($pack_info->{n_doubles});
    $size += $s;
    $n_elts += $e;
    print "double size: $s $e\n";

    for $gp (@{$pack_info->{global_pointers}})
    {
	($s, $e) = $buf->sizeof_global_pointer($gp);
	print "gp $gp size: $s $e\n";
	$size += $s;
	$n_elts += $e;
    }

    return ($size, $n_elts);
}

sub pack_buffer
{
    my($buf, $pack_info) = @_;
    my($item, $ref, $type, $count);

    for $item (@{$pack_info->{items}})
    {
	($type, $count, $ref) = @$item;

	length($type) == 1 or die;
	print "put_string '$type'\n" if $debug;
	$buf->put_string($type);

	print "put_int $count\n" if $debug;
	$buf->put_int($count);

	if ($type eq 'i')
	{
	    $buf->put_int($$ref);
	    print "put_int $$ref\n" if $debug;
	}
	elsif ($type eq 'd')
	{
	    $buf->put_double($$ref);
	    print "put_double $$ref\n" if $debug;
	}
	elsif ($type eq 'c')
	{
	    $buf->put_string($$ref);
	    length($$ref) == $count or die;
	    print "put_char '$$ref'\n" if $debug;
	}
	elsif ($type eq 'g')
	{
	    $buf->put_global_pointer($ref);
	    print "put_gp $ref\n" if $debug;
	}
	elsif ($type eq 'r')
	{
	    my $refname = $ref->type();
	    my $gp = $ref->gp();
	    $buf->put_global_pointer($gp);
	    print "put_gp $refname $ref\n" if $debug;

	    my $l = length($refname);
	    $buf->put_int($l);
	    print "put_int $l\n" if $debug;
	    $buf->put_string($refname);
	    print "put_string '$refname'\n" if $debug;
	}
    }
}

sub unpack_buffer
{
    my($buf, $indent) = @_;
    my($rc, $type, $count, $elt, $key, $val, $i);

    my $spaces = ' ' x ($indent * 2);

    $buf->get_string(\$type, 1);
    print $spaces, "got type $type\n" if $debug;
    
    $buf->get_int(\$count);
    print $spaces, "got count $count\n" if $debug;
    
    if ($type eq 'u')
    {
	$count == 1 or die "Nperl can't do arrays of undefs yet\n";
	$rc = undef;
    }
    elsif ($type eq 'i')
    {
	$count == 1 or die "Nperl can't do arrays of ints yet\n";
	$buf->get_int(\$rc);
	print $spaces, "got int $rc\n" if $debug;
    }
    elsif ($type eq 'c')
    {
	$buf->get_string(\$rc, $count);
	print $spaces, "got string $rc\n" if $debug;
    }
    elsif ($type eq 'd')
    {
	$count == 1 or die "Nperl can't do arrays of doubles yet\n";
	$buf->get_double(\$rc);
	print $spaces, "got double $rc\n" if $debug;
    }
    elsif ($type eq 'h')
    {
	$rc = {};
	for ($i = 0; $i < $count; $i++)
	{
	    $key = &unpack_buffer($buf, $indent + 1);
	    $val = &unpack_buffer($buf, $indent + 1);
	    $rc->{$key} = $val;
	}
    }
    elsif ($type eq 'a')
    {
	$rc = [];
	for ($i = 0; $i < $count; $i++)
	{
	    $elt = &unpack_buffer($buf, $indent + 1);
	    push(@$rc, $elt);
	}
    }
    elsif ($type eq 'g')
    {
	$buf->get_global_pointer(\$rc);
	print "got gp $rc\n" if $debug;
    }
    elsif ($type eq 'r')
    {
	my($gp, $name, $namelen);

	$buf->get_global_pointer(\$gp);
	$buf->get_int(\$namelen);
	$buf->get_string(\$name, $namelen);
	print "** got rref $name $gp\n" if $debug;

	#
	# If we have a local gp, then we return the reference.
	# The global pointer is pointing to the RRT, so we
	# localize and invoke the ref method.
	#
	# Otherwise, construct a RemoteReference and return that.
	#
	if ($gp->is_local())
	{
	    print "GP local\n" if $debug;

	    my $ref = $gp->dereference;
	    print "Deref returns $ref\n" if $debug;

	    $rc = $ref->ref();

	    print "Ref is $rc\n" if $debug;
	}
	else
	{
	    $rc = new Nexus::RemoteReference($gp, $name);
	    print "Created remoteref $rc\n" if $debug;
	}
    }
    
    else
    {
	die "Unknown type $type\n";
    }
    return $rc;
}

sub dump_packed
{
    my($pack_info) = @_;
    my($item, $ref, $type, $count, $key);

    for $key (qw(n_ints n_chars n_doubles))
    {
	print "$key = $pack_info->{$key}\n";
    }

    for $item (@{$pack_info->{items}})
    {
	($type, $count, $ref) = @$item;
	print "    $type $count $ref ";
	print $$ref if ref($ref) eq "SCALAR";
	print "\n";
    }
}

sub pack_item
{
    my($item, $pack_info) = @_;
    my($ref);

    $ref = ref($$item);
    if ($ref eq "ARRAY")
    {
	&pack_array($$item, $pack_info);
    }
    elsif ($ref eq "HASH")
    {
	&pack_hash($$item, $pack_info);
    }
    elsif ($ref eq "Nexus::GlobalPointer")
    {
	&pack_global_pointer($$item, $pack_info);
    }
    elsif ($ref eq "Nexus::RemoteReference")
    {
	&pack_remote_reference($$item, $pack_info);
    }
    elsif ($ref ne '')
    {
	&pack_ref($$item, $pack_info);
    }
    elsif (!defined($$item))
    {
#	print "Got undef!\n";
	&pack_undef($pack_info);
    }
    elsif (Nexus::isa_string($item))
    {
	&pack_string($item, $pack_info);
    }
    elsif (Nexus::isa_int($item))
    {
	&pack_int($item, $pack_info);
    }
    elsif (Nexus::isa_double($item))
    {
	&pack_double($item, $pack_info);
    }
    else
    {
	die "Unknown type for item <$item> <$$item> <$ref>\n";
    }
}

sub pack_global_pointer
{
    my($gp, $pack_info) = @_;

    push(@{$pack_info->{items}}, ['g', 1, $gp]);
    $pack_info->add_char(1);
    $pack_info->add_int(1);
    $pack_info->add_gp($gp);
}

#
# Remote reference is packed as name & global pointer.
#
sub pack_remote_reference
{
    my($rref, $pack_info) = @_;
    my($gp, $refname);

    $gp = $rref->gp();
    $refname = $rref->type();

    print "packing rref $refname $gp\n" if $debug;
    push(@{$pack_info->{items}}, ['r', 1, $rref]);
    $pack_info->add_char(1);
    $pack_info->add_int(1);
    $pack_info->add_gp($gp);

    #
    # RRefs send along the type name. We don't use pack_string
    # here.
    #

    $pack_info->add_char(length($refname));
    $pack_info->add_int(1);
}

#
# Pack an arbitrary reference.
#
# This is done by creating a remote reference trap for this ref, 
# and sending along the trap's global pointer using the same
# protocol as pack_remote_reference.
#
# This is easy as RRT has the same signature as RemoteReference. :-)
#
sub pack_ref
{
    my($ref, $pack_info) = @_;
    my($rrt);

    $rrt = new Nexus::RRT($ref);

    &pack_remote_reference($rrt, $pack_info);
}

sub pack_int
{
    my($item, $pack_info) = @_;

    push(@{$pack_info->{items}}, ['i', 1, $item]);

    $pack_info->add_char(1);
    $pack_info->add_int(1);
    $pack_info->add_int(1);
}

sub pack_undef
{
    my($pack_info) = @_;

    push(@{$pack_info->{items}}, ['u', 1]);
    $pack_info->add_char(1);
    $pack_info->add_int(1);
}

sub pack_double
{
    my($item, $pack_info) = @_;

    push(@{$pack_info->{items}}, ['d', 1, $item]);
    $pack_info->add_char(1);
    $pack_info->add_int(1);
    $pack_info->add_double(1);
}

sub pack_string
{
    my($item, $pack_info) = @_;
    my($len) = length($$item);
    
    push(@{$pack_info->{items}}, ['c', $len, $item]);

    $pack_info->add_char(1);
    $pack_info->add_char($len);
    $pack_info->add_int(1);
}

sub pack_array
{
    my($array, $pack_info) = @_;
    my($item, $len, $i);

    $len = $#$array + 1;

    push(@{$pack_info->{items}}, ['a', $len, undef]);
    $pack_info->add_char(1);
    $pack_info->add_int(1);

    for ($i = 0; $i < $len; $i++)
    {
	&pack_item(\$$array[$i], $pack_info);
    }
}

sub pack_hash
{
    my($hash, $pack_info) = @_;
    my($item, $len, @keys, $key, $elt);

    @keys = keys(%$hash);
    $len = @keys;

    push(@{$pack_info->{items}}, ['h', $len, undef]);
    $pack_info->add_char(1);
    $pack_info->add_int(1);

    for $key (@keys)
    {
	print "key=$key\n" if $debug;
	&pack_item(\$key, $pack_info);
	&pack_item(\$hash->{$key}, $pack_info);
    }
}

package Nexus::DeepPacker::PackInfo;

sub new
{
    my($class, $buf) = @_;
    my($pack_info);

    $pack_info = {
	items => [],
	n_ints => 0,
	n_chars => 0,
	n_doubles => 0,
	n_global_pointers => 0,
	size => 0,
	n_elts => 0,
	buf => $buf
    };
    bless ($pack_info, $class);
    return $pack_info;
}

sub DESTROY
{
    my($self) = @_;
    print "DESTROY pack_info $self\n" if $debug;
}

sub add_int
{
    my($self, $n) = @_;
    my($s, $e);

#    $n == 1 or die "nperl only groks put_int(1)";

    ($s, $e) = $self->{buf}->sizeof_int(1);
    $self->{size} += $s;
    $self->{n_elts} += $e;
}

sub add_double
{
    my($self, $n) = @_;
    my($s, $e);

#    $n == 1 or die "nperl only groks put_double(1)";

    ($s, $e) = $self->{buf}->sizeof_double(1);
    $self->{size} += $s;
    $self->{n_elts} += $e;
}

sub add_char
{
    my($self, $n) = @_;
    my($s, $e);

    ($s, $e) = $self->{buf}->sizeof_char($n);
    $self->{size} += $s;
    $self->{n_elts} += $e;
}

sub add_gp
{
    my($self, $gp) = @_;
    my($s, $e);

    ($s, $e) = $self->{buf}->sizeof_global_pointer($gp);
    $self->{size} += $s;
    $self->{n_elts} += $e;
}


1;
