package Nexus::RRT;
require Carp;

my %ref2rrts = ();
my %rrt2refs = ();

sub dump_rrts
{
    &dump_all_rrts;
}

sub dump_active_rrts
{
    my($ref, $rrt);

    print "Dumping ", int(keys(%ref2rrts)), " RRTS\n";
    while (($ref, $rrt) = each(%ref2rrts))
    {
	print "ref=$ref rrt=$rrt\n";
	Nexus::sv_dump($rrt);
    }

}

sub dump_all_rrts
{
    my($ref, $rrt);

    print "Dumping ", int(@allrrts), " RRTS\n";
    for $rrt (@allrrts)
    {
	print "rrt=$rrt\n";
	Nexus::sv_dump($rrt);
    }

}

@allrrts = ();

sub boot
{
#    Nexus::register_handler("Nexus::RRT::incr", "incr",
#			    method_call => 1,
#			    unpack_buffer => 1);
#    Nexus::register_handler("Nexus::RRT::decr", "decr", 
#			    method_call => 1,
#			    unpack_buffer => 1);
}

sub new
{
    my($class, $ref) = @_;
    my($self);

    print "RRT::new <$class> <$ref>\n" if $debug > 1;
    
    if ($self = $ref2rrts{$ref})
    {
	print "Found existing rrt $self for $ref\n" if $debug > 1;
	$self->[1] = $ref;
	$self->set_type;
	return $self;
    }
    
    $self = [0, $ref, undef];

    bless $self, $class;

    $self->[2] = Nexus::GlobalPointer->new($self);

    $self->set_type;

    $ref2rrts{$ref} = $self;
#    $rrt2refs{$self} = "$ref";

#    push(@allrrts, $self);

    return $self;
}

#
# Set the type name for this ref.
#
sub set_type
{
    my($self) = @_;
    
    $self->[3] = ref($self->[1]);
#    print "Set type to <$self->[3]>\n";
}

sub type
{
    my($self) = @_;
    return $self->[3];
}

sub gp
{
    my($self) = @_;
#    print "In RRT:gp self=$self gp=$self->[2] caller=", join(":", caller), "\n";
    return $self->[2];
}

sub ref
{
    my($self) = @_;
    return $self->[1];
}

sub incr_handler
{
    my($self, $buf) = @_;
    my($count);

    ($count) = $buf->unpack();

    $self->incr($count);
}

sub incr
{
    my($self, $count) = @_;

    $count = 1 unless defined($count);
    
    print "INCR $self $self->[0] $self->[1] $count\n" if $debug;
    $self->[0] += $count;
}

sub decr_handler
{
    my($self, $buf) = @_;
    my($count);

    ($count) = $buf->unpack();

    $self->decr($count);
}

sub decr
{
    my($self, $count) = @_;

    $count = 1 unless defined($count);

    print "DECR $self $self->[0] $self->[1] $count\n" if $debug;
    
    $self->[0] -= $count;

    if ($self->[0] <= 0)
    {
	my($ref) = "$self->[1]";
	print "Destroy $ref\n" if $debug;
	if ($debug > 1)
	{
	    Nexus::sv_dump($self->[1]);
	}
	undef $self->[1];
	print "Before del\n" if $debug > 2;
	Nexus::sv_dump($self) if $debug > 2;
	delete $ref2rrts{$ref}; # XXX
	print "after del\n" if $debug > 2;
	Nexus::sv_dump($self) if $debug > 2;
    }
}

sub print
{
    my($self) = @_;

    print "$self: $self->[0] $self->[1]\n";
}

sub DESTROY
{
    my($self) = @_;
    my($refnam);

#    $refnam = $rrt2refs{$self};

    print "RRT::DESTROY $self $self->[0] $self->[1] <$refnam>\n" if $debug;
#    print "delete <$rrt2refs{$self}> <$ref2rrts{$refnam}>\n";;
#    delete $rrt2refs{$self};
#    delete $ref2rrts{$refnam};
}

1;
