
package Nexus::RemoteReference;

$OVERLOAD{'""'} = 'as_string';
$OVERLOAD{fallback} = 1;

sub AUTOLOAD
{
    my($self, @args) = @_;
    my($func) = $AUTOLOAD;
    $func =~ s/^Nexus::RemoteReference:://;
    print "RRef autoload: $func ", $self->type(), " args=@args\n";
}
     
use Carp;

sub boot
{
}

sub new
{
    my($class, $gp, $type) = @_;
    my($self, $buf);
    
    $self = [$gp, $type];
    print "$class :: new $gp $type\n" if $debug;

    if ($type =~ /URI::URL/)
    {
	print "ACK! passed a remote ref to $type\n";
	&main::traceback();
    }

    if (ref($gp) ne "Nexus::GlobalPointer")
    {
	die "Nexus::RemoteReference::new got a bogus gp $gp\n";
    }

    bless $self, $class;

    $buf = Nexus::init_rsr($gp, "Nexus::RRT::incr");
    my($s, $e) = $buf->sizeof_int(1);
    $buf->set_size($s, $e);
    $buf->put_int(1);
    $buf->send_rsr();
    Nexus::poll(0);
    return $self;
}

sub test
{
    my($self) = @_;

    return $self->gp()->test();
}

sub as_string
{
    my($self) = @_;

    return "RREF($$self[0],$$self[1])";
}

sub gp
{
    return ${$_[0]}[0];
}

sub type
{
    return ${$_[0]}[1];
}

sub DESTROY
{
    my($self) = @_;
    my($gp, $buf, $ref, $type);

    ($gp, $type) = @$self;

    print "RR::DESTROY $self gp=<$gp> type=<$type> <", ref($gp), ">\n" if $debug;

    if (($ref = ref($gp)) ne "Nexus::GlobalPointer")
    {
	&main::traceback();
	die "Nexus::RemoteReference has invalid gp \"$gp\" with ref \"$ref\"\n";
    }

    $buf = Nexus::init_rsr($gp, "Nexus::RRT::decr");
    my($s, $e) = $buf->sizeof_int(1);
    $buf->set_size($s, $e);
    $buf->put_int(1);

    print "send rsr to $gp\n" if $debug;
    $buf->send_rsr();
    #
    # This is causing problems with looping. Ripped out and things
    # work btter (and stop leaking memory)
#    Nexus::poll(0);
}


1;

__END__
