#include "EXTERN.h"
#include "perl.h"
#include "nexus.h"
#include "pnx.h"

/*#undef PNX_DEBUG*/
/*#define PNX_DEBUG 1*/

#define NEXUS_IMMEDIATE_HANDLERS

#define NPERL_DEFAULT_HANDLER "NPerlDefaultHandler"

static void shutdown_nodes(nexus_node_t *nodes, int n_nodes);
static int new_process_params(char *buf, int size);
static void usage_message();

void NexusExit();
void NexusUnknownHandler(void *address, nexus_buffer_t *buffer,
			 char *handler_name, int handler_id);
static void shutdown_node_handler(void *address, nexus_stashed_buffer_t *buffer);
#define     SHUTDOWN_NODE_HANDLER_HASH		196

static void shutdown_node_reply_handler(void *address, nexus_buffer_t *buffer);
#define     SHUTDOWN_NODE_REPLY_HANDLER_HASH	847

static void rrt_incr_handler(void *address, nexus_buffer_t *buffer);
#define	    RRT_INCR_HANDLER_NAME "Nexus::RRT::incr"
#define     RRT_INCR_HANDLER_HASH 418

static void rrt_decr_handler(void *address, nexus_buffer_t *buffer);
#define	    RRT_DECR_HANDLER_NAME "Nexus::RRT::decr"
#define     RRT_DECR_HANDLER_HASH 404

static void call_user_rsr(pnx_rsr_t *rsr);
static void call_decr_rsr(pnx_rsr_t *rsr);
static void call_incr_rsr(pnx_rsr_t *rsr);

static nexus_handler_t system_handlers[] =
{ \

      /*
      {"NexusExit",
      NEXUS_EXIT_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) NexusExit},
  {"NexusUnknownHandler",
       NEXUS_UNKNOWN_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) NexusUnknownHandler},
  {"NexusAcquiredAsNode",
       NEXUS_ACQUIRED_AS_NODE_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) NexusAcquiredAsNode},
  {"shutdown_node_handler",
       SHUTDOWN_NODE_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_THREADED,
       (nexus_handler_func_t) shutdown_node_handler},
  {"shutdown_node_reply_handler",
       SHUTDOWN_NODE_REPLY_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) shutdown_node_reply_handler},
*/
  {(char *) NULL,
       0,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) NULL},
};

static nexus_node_t *initial_node_list;
static int initial_node_count;

static void pnx_rsr_handler(void *address, nexus_buffer_t *buffer,
			    char *handler_name, int handler_id);

static nexus_handler_t pnx_handlers[] =
{ \
  {"NexusUnknownHandler",
       NEXUS_UNKNOWN_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) pnx_rsr_handler},
  {RRT_INCR_HANDLER_NAME,
       RRT_INCR_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) rrt_incr_handler},
  {RRT_DECR_HANDLER_NAME,
       RRT_DECR_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) rrt_decr_handler},
  {(char *) NULL,
       0,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) NULL},
};

static pnx_rsr_queue_t *pnx_rsr_q;

/*
 * Stuff for select() style notification of pending messages.
 */
static int notify_r_fd, notify_w_fd;
static FILE *notify_fp;
static nexus_mutex_t notify_lock;

static SV *fd_add_notifier = (SV *) NULL;
static SV *fd_remove_notifier = (SV *) NULL;
static SV *fd_add_notifier_arg = (SV *) NULL;
static SV *fd_remove_notifier_arg = (SV *) NULL;

static void fd_add_notifier_func(int);
static void fd_remove_notifier_func(int);

static SV *fd_saved_notifier_fds[1024];

/*
 * This is the thread that will run the Perl interpreter.
 *
 * Note that the interpreter has already been created (since
 * we now always enter Nperl through a standard Perl main
 * routine.
 *
 * Thus all we have to do is invoke the perl routine main::NexusNodeBoot.
 */
static void *pnx_perl_interpreter_thread(void *user_arg)
{
    dSP ;
    PUSHMARK(sp) ;
    PUTBACK ;
    printf("Invoking main::NexusNodeBoot\n");
    perl_call_pv("main::NexusNodeBoot", G_DISCARD);
    printf("main::NexusNodeBoot returns\n");
}

/*
 * NexusBoot()
 */
int NexusBoot()
{
    int rc;
#ifdef NX_DEBUG_LEVEL
    extern int _nx_debug_level;

    _nx_debug_level = NX_DEBUG_LEVEL;
#endif
/*    nexus_printf("NexusBoot(): entering\n");*/
    
    nexus_register_handlers(system_handlers);

    pnx_boot();
    /*
    pnx_perl_boot();
    pnx_sr_boot();
    */
    pnx_mb_boot();
    pnx_smb_boot();
    pnx_gp_boot();
    pnx_attach_boot();

    do {
	dSP ;
	PUSHMARK(sp) ;
	PUTBACK ;
/*	printf("Invoking Nexus::boot\n");*/
	perl_call_pv("Nexus::boot", G_DISCARD | G_EVAL);
/*	printf("Nexus::boot returns\n");*/
    } while(0);

#ifndef NEXUS_LITE
    /*
     * If we're note the master node, we need to create a thread
     * to run the perl interpreter. The thread will immediately
     * invoke main::NexusNodeBoot which should spin on Nexus::poll.
     */
    if (!nexus_master_node())
    {
	nexus_thread_t thread;
	
	nexus_thread_create(&thread, (nexus_thread_attr_t *) NULL,
			    pnx_perl_interpreter_thread, NULL);
    }

#endif /* NEXUS_LITE */
    rc = 0;
    
/*    nexus_printf("NexusBoot(): exiting, returning %d\n", rc);*/

    return (rc);
} /* NexusBoot() */

void pnx_boot()
{
    nexus_mutex_init(&notify_lock, (nexus_mutexattr_t *) NULL);

    notify_r_fd = -1;
    notify_w_fd = -1;
    notify_fp = (FILE *) NULL;
    
    pnx_rsrq_init();
    nexus_register_handlers(pnx_handlers);

    bzero(fd_saved_notifier_fds, sizeof(fd_saved_notifier_fds));
}

FILE *pnx_get_notify_fh()
{
    nexus_mutex_lock(&notify_lock);
    if (notify_fp == (FILE *) NULL)
    {
	int p[2];
	pipe(p);
	notify_r_fd = p[0];
	notify_w_fd = p[1];

	notify_fp = fdopen(notify_r_fd, "r");
    }
    nexus_mutex_unlock(&notify_lock);
    return notify_fp;
}

void pnx_register_fd_notifiers(SV *add_ref, SV *add_arg,
			       SV *remove_ref, SV *remove_arg)
{
    SV *add, *remove;
/*    printf("register fd notifiers\n");*/

    if (SvROK(add_ref))
    {
	add = SvRV(add_ref);
    }
    else
    {
	croak("register_fd_notifiers needs code ref");
    }

    if (SvROK(remove_ref))
    {
	remove = SvRV(remove_ref);
    }
    else
    {
	croak("register_fd_notifiers needs code ref");
    }

    fd_add_notifier = add;
    fd_add_notifier_arg = add_arg;
    fd_remove_notifier = remove;
    fd_remove_notifier_arg = remove_arg;

    SvREFCNT_inc(add_arg);
    SvREFCNT_inc(remove_arg);

    SvREFCNT_inc(add);
    SvREFCNT_inc(remove);

    nexus_register_fd_notifiers(fd_add_notifier_func, fd_remove_notifier_func);
}

static void fd_add_notifier_func(int fd)
{
    dSP;
    GV *gv;
    SV *sv;
    FILE *fp;
    
    
/*    printf("C got add %d\n", fd);*/
/*    pnx_sv_dump(fd_add_notifier_arg);*/

    if ((fp = fdopen(fd, "r")) == (FILE *) NULL)
    {
	croak("fd_add_notifier_func: fdopen failed");
    }
    setbuf(fp, NULL);
    gv = newGVgen("Nexus");
    if (!do_open(gv, "<&", 2, fp))
    {
	croak("fd_add_notifier_func: do_open failed");
    }

    sv = sv_bless(newRV((SV *)gv), gv_stashpv("Nexus", 1));

    SvREFCNT_inc(sv);

    fd_saved_notifier_fds[fd] = sv;

    PUSHMARK(sp);
    XPUSHs(fd_add_notifier_arg);
    XPUSHs(sv_2mortal(sv));
    PUTBACK;
    perl_call_sv(fd_add_notifier, G_EVAL | G_DISCARD);
}

static void fd_remove_notifier_func(int fd)
{
    dSP;
/*    printf("C got remove %d\n", fd);*/
    PUSHMARK(sp);
    XPUSHs(fd_remove_notifier_arg);
    XPUSHs(fd_saved_notifier_fds[fd]);

    PUTBACK;
    perl_call_sv(fd_remove_notifier, G_EVAL | G_DISCARD);

    SvREFCNT_dec(fd_saved_notifier_fds[fd]);
    fd_saved_notifier_fds[fd] = NULL;
}

void pnx_set_initial_node_list(nexus_node_t *initial_nodes,
			       int n_initial_nodes)
{
    initial_node_list = initial_nodes;
    initial_node_count = n_initial_nodes;
}

void pnx_get_initial_node_list(nexus_node_t **list, int *count)
{
    *list = initial_node_list;
    *count = initial_node_count;
}

SV *pnx_create_context(SV *gpref, char *pathname, int size)
{
#ifdef NEXUS_LITE
    croak("Nexus::create_context() not defined in NexusLite");
#else
    int rc;
    nexus_global_pointer_t *gp, *new_gp;

    gp = pnx_gp_get_gp(gpref, "pnx_create_context");

    NexusMalloc(pnx_create_context, new_gp, nexus_global_pointer_t *,
		sizeof(nexus_global_pointer_t));

    nexus_create_context(gp, pathname, size, new_gp, &rc,
			 (nexus_create_context_handle_t *) NULL);

    if (rc != 0)
    {
	printf("create_context failed with rc=%d\n", rc);
	return &sv_undef;
    }

/*    nexus_printf("create_context returns rc=%d\n", rc); */

    gpref = sv_setref_iv(newSVsv(&sv_undef), "Nexus::GlobalPointer", 
			 (IV) new_gp);
    return gpref;
#endif
}

/*
 * Destroy the current context. This should cause all the perl
 * interps to exit nicely...
 */
void pnx_destroy_current_context()
{
#ifndef NEXUS_LITE
    PerlInterpreter *interp;

    interp = pnx_perl_get_interpreter();
    pnx_perl_destroy(interp);
    pnx_perl_free(interp);
#endif
    nexus_destroy_current_context(0);
}

SV *pnx_init_rsr(SV *gp_ref, char *subroutine)
{
    nexus_global_pointer_t *gp;
    pnx_buffer_t *buffer;
    SV *buf_val_ref;

    NexusMalloc(pnx_init_rsr, buffer, pnx_buffer_t *,
		sizeof(pnx_buffer_t));

    strcpy(buffer->subname, subroutine);

    if (sv_isa(gp_ref, "Nexus::GlobalPointer"))
    {
	gp = pnx_gp_get_gp(gp_ref, "pnx_init_rsr");
    }
    else if (sv_isa(gp_ref, "Nexus::RemoteReference"))
    {
	gp_ref = pnx_rrt_gp(gp_ref);
	gp = pnx_gp_get_gp(gp_ref, "pnx_init_rsr");
    }
    else
    {
	croak("Must pass global pointer or remote reference to init_rsr");
    }

#ifdef PNX_DEBUG
    printf("init_rsr sub=%s gp=%s\n", subroutine,
	   nexus_global_pointer_string(gp));
#endif

    nexus_init_remote_service_request(&buffer->buffer, gp,
				      buffer->subname,
				      nexus_handler_hash(buffer->subname));
    
    buf_val_ref = sv_setref_iv(newSVsv(&sv_undef),
			       "Nexus::MessageBuffer", (IV) buffer);
    return buf_val_ref;
}

static void rrt_incr_handler(void *address, nexus_buffer_t *buffer)
{
    pnx_rsr_t *rsr;

#ifdef PNX_DEBUG
    printf("In rrt_incr_handler address=%x\n", address);
#endif

    NexusMalloc(pnx_rsr_handler, rsr, pnx_rsr_t *, sizeof(pnx_rsr_t));

    rsr->func = call_incr_rsr;
    rsr->address = (SV *) address;

    nexus_get_int(buffer, &rsr->u.count, 1);

#ifdef NEXUS_IMMEDIATE_HANDLERS

    (rsr->func)(rsr);

#else
    
    pnx_rsrq_put(pnx_rsr_q, rsr);
    nexus_mutex_lock(&notify_lock);
/*    printf("Would write to notify_fd %d\n", notify_w_fd);*/
    if (notify_w_fd != -1)
    {
	char one = 1;
	write(notify_w_fd, &one, 1);
    }
    nexus_mutex_unlock(&notify_lock);

#endif
}

static void rrt_decr_handler(void *address, nexus_buffer_t *buffer)
{
    pnx_rsr_t *rsr;

#ifdef PNX_DEBUG
    printf("In rrt_decr_handler address=%x\n", address);
#endif

    NexusMalloc(pnx_rsr_handler, rsr, pnx_rsr_t *, sizeof(pnx_rsr_t));

    rsr->func = call_decr_rsr;
    rsr->address = (SV *) address;

    nexus_get_int(buffer, &rsr->u.count, 1);

#ifdef NEXUS_IMMEDIATE_HANDLERS

    (rsr->func)(rsr);

#else
    
    pnx_rsrq_put(pnx_rsr_q, rsr);
    nexus_mutex_lock(&notify_lock);
/*    printf("Would write to notify_fd %d\n", notify_w_fd);*/
    if (notify_w_fd != -1)
    {
	char one = 1;
	write(notify_w_fd, &one, 1);
    }
    nexus_mutex_unlock(&notify_lock);

#endif
}

static void pnx_rsr_handler(void *address, nexus_buffer_t *buffer,
			    char *handler_name, int handler_id)
{
    pnx_rsr_t *rsr;

#ifdef PNX_DEBUG
    printf("In pnx_rsr_handler address=%x name=%s\n", 
	   address, handler_name);
#endif

    NexusMalloc(pnx_rsr_handler, rsr, pnx_rsr_t *, sizeof(pnx_rsr_t));

    rsr->func = call_user_rsr;
    rsr->address = (SV *) address;
    nexus_stash_buffer(buffer, &rsr->u.user.buffer);
    NexusMalloc(pnx_rsr_handler, rsr->u.user.subname,
		char *, strlen(handler_name) + 1);
    strcpy(rsr->u.user.subname, handler_name);

#ifdef NEXUS_IMMEDIATE_HANDLERS
    
    (rsr->func)(rsr);
    
#else

    pnx_rsrq_put(pnx_rsr_q, rsr);

    nexus_mutex_lock(&notify_lock);
/*    printf("Would write to notify_fd %d\n", notify_w_fd);*/
    if (notify_w_fd != -1)
    {
	char one = 1;
	write(notify_w_fd, &one, 1);
    }
    nexus_mutex_unlock(&notify_lock);
#endif
}

static void call_incr_rsr(pnx_rsr_t *rsr)
{
    SV *sv;
    
    sv = newRV(rsr->address);

    pnx_rrt_incr(sv, rsr->u.count);

    SvREFCNT_dec(sv);

    NexusFree(rsr);
}

static void call_decr_rsr(pnx_rsr_t *rsr)
{
    SV *sv;
    
    sv = newRV(rsr->address);

    pnx_rrt_decr(sv, rsr->u.count);

    SvREFCNT_dec(sv);

    NexusFree(rsr);
}

static void call_user_rsr(pnx_rsr_t *rsr)
{
    SV *buf_ref;
    SV *addr_ref;
    SV *handler_name;
    int default_handler;

#ifdef PNX_DEBUG
    printf("call_user_rsr address=%x name=%s\n",
	   rsr->address, rsr->u.user.subname);
#endif

    if (rsr->address == NULL)
    {
	addr_ref = newSVsv(&sv_undef);
    }
    else if (!SvOBJECT(rsr->address))
    {
	addr_ref = newRV(rsr->address);
    }
    else
    {
	char *name;

	name = HvNAME(SvSTASH(rsr->address));
/*	printf("got name=%s\n", name);*/
	if (strcmp(name, "Nexus::RRT") == 0)
	{
	    /*
	     * The ref is $$self[1] in the RRT. Just yank it out.
	     */
	    addr_ref = *av_fetch((AV *) rsr->address, 1, 0);
	}
	else if (strcmp(name, "Nexus::GlobalPointer") == 0)
	{
	    addr_ref = pnx_gp_dereference(rsr->address);
	    addr_ref = sv_2mortal(addr_ref);
	}
	/*
	 * Hmm. The natural thing is to just return a ref to the
	 * object, since that's how the GP got created in the first place.
	 *
	else
	{
	    pnx_sv_dump(rsr->address);
	    croak("Unknown type %s in pnx_call_rsr %s:%d",
		  name, __FILE__, __LINE__);
	}
	 */
	else
	{
	    addr_ref = newRV(rsr->address);
	}
    }

    buf_ref = sv_setref_iv(newSVsv(&sv_undef),
			   "Nexus::StashedMessageBuffer",
			   (IV) rsr);

    handler_name = newSVpv(rsr->u.user.subname, strlen(rsr->u.user.subname));

    /*
     * Invoke Nexus::invoke_handler(handler_name, address, buffer)
     */

    {
	dSP;
	PUSHMARK(sp) ;
	/*
	 * Important: NO sv_2mortal on the addr_ref...
	 */
	XPUSHs(sv_2mortal(handler_name));
	XPUSHs(addr_ref);
	XPUSHs(sv_2mortal(buf_ref));
	PUTBACK ;

	perl_call_pv("Nexus::invoke_handler", G_EVAL | G_DISCARD);
    }
	
}

void pnx_poll(int block)
{

#ifdef NEXUS_IMMEDIATE_HANDLERS

  if (block)
	nexus_blocking_poll();
    else
	nexus_poll();
    
#else

    pnx_rsr_t *rsr;

    nexus_mutex_lock(&pnx_rsr_q->lock);

#ifdef PNX_DEBUG
    printf("pnx_poll(): block=%d polling=%d\n", block, pnx_rsr_q->polling);
#endif
    if (pnx_rsr_q->polling)
    {
#ifdef PNX_DEBUG
	printf("pnx_poll(): returning due to polling\n");
#endif
	nexus_mutex_unlock(&pnx_rsr_q->lock);
	return;
    }
    pnx_rsr_q->polling = 1;

    if (pnx_rsr_q->first.next == &pnx_rsr_q->first)
    {
	if (!block)
	{
#ifdef PNX_DEBUG
	    printf("pnx_poll(): returning due to block==0 and empty\n");
#endif
	    pnx_rsr_q->polling = 0;
	    nexus_mutex_unlock(&pnx_rsr_q->lock);
	    return;
	}
	else
	{
	    pnx_rsr_q->waiting++;
	    while (pnx_rsr_q->first.next == &pnx_rsr_q->first)
	    {
#ifdef PNX_DEBUG
		printf("pnx_poll(): waiting for message\n");
#endif
#ifdef NEXUS_LITE
		nexus_mutex_unlock(&pnx_rsr_q->lock);
		nexus_blocking_poll();
		nexus_mutex_lock(&pnx_rsr_q->lock);
#endif
		nexus_cond_wait(&pnx_rsr_q->cond, &pnx_rsr_q->lock);
	    }
#ifdef PNX_DEBUG
	    printf("pnx_poll(): got message\n");
#endif
	    pnx_rsr_q->waiting--;
	}
    }

    while (pnx_rsr_q->first.next != &pnx_rsr_q->first)
    {
	rsr = pnx_rsr_q->first.next;
	pnx_rsr_q->first.next = rsr->next;

	if (rsr == pnx_rsr_q->last)
	    pnx_rsr_q->last = &pnx_rsr_q->first;
	
	nexus_mutex_unlock(&pnx_rsr_q->lock);
#ifdef PNX_DEBUG
	printf("pnx_poll(): invoking rsr %x\n", rsr);
#endif
	(rsr->func)(rsr);
#ifdef PNX_DEBUG
	printf("pnx_poll(): invoked rsr %x\n", rsr);
#endif
	nexus_mutex_lock(&pnx_rsr_q->lock);
    }
    
    pnx_rsr_q->polling = 0;
#ifdef PNX_DEBUG
    printf("pnx_poll(): normal return\n");
#endif
    nexus_mutex_unlock(&pnx_rsr_q->lock);
#endif /* NEXUS_LITE */
}

void pnx_rsrq_init()
{
    NexusMalloc(sr_rsrq_init, pnx_rsr_q, pnx_rsr_queue_t *,
		sizeof(pnx_rsr_queue_t));

    pnx_rsr_q->polling = 0;
    pnx_rsr_q->waiting = 0;
    pnx_rsr_q->last = &pnx_rsr_q->first;
    pnx_rsr_q->first.next = &pnx_rsr_q->first;

    nexus_mutex_init(&pnx_rsr_q->lock, (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&pnx_rsr_q->cond, (nexus_condattr_t *) NULL);
}

int pnx_rsrq_empty(pnx_rsr_queue_t *q)
{
    int rc;
    nexus_mutex_lock(&q->lock);
    if (q->first.next == &q->first)
    {
	rc = 1;
    }
    else
    {
	rc = 0;
    }
    
    nexus_mutex_unlock(&q->lock);
    return rc;
}

pnx_rsr_t *pnx_rsrq_get(pnx_rsr_queue_t *q, int wait_flag)
{
    pnx_rsr_t *msg;

    nexus_mutex_lock(&q->lock);

    if (q->first.next == &q->first)
    {
	if (wait_flag == PNX_DONT_WAIT)
	{
	    nexus_mutex_unlock(&q->lock);
	    return (pnx_rsr_t *) NULL;
	}
	else
	{
	    q->waiting++;
	    while (q->first.next == &q->first)
	    {
/*		nexus_printf("waiting for message\n"); */
		nexus_cond_wait(&q->cond, &q->lock);
	    }
/*	    nexus_printf("got message\n"); */
	    q->waiting--;
	}
    }

    msg = q->first.next;
    q->first.next = msg->next;

    if (msg == q->last)
	q->last = &q->first;

    nexus_mutex_unlock(&q->lock);
    return msg;
}

void pnx_rsrq_put(pnx_rsr_queue_t *q, pnx_rsr_t *msg)
{
    nexus_mutex_lock(&q->lock);
    msg->next = &q->first;
    q->last->next = msg;
    q->last = msg;
    if (q->waiting)
    {
	nexus_cond_signal(&q->cond);
    }
    nexus_mutex_unlock(&q->lock);
}

