#include "EXTERN.h"
#include "perl.h"
#include "nexus.h"
#include "pnx.h"

void pnx_gp_boot(void)
{
}

/*
 * Return a GlobalPointer to the given perl variable.
 */
SV *pnx_gp_new(char *clas, SV *perlvar)
{
    int ival;
    double dval;
    char *sval;
    SV *gpv, *gpref, *stash;
    nexus_global_pointer_t *gp;

    if (!SvROK(perlvar))
    {
	croak("Nexus::GlobalPointer::new wants a reference");
	return &sv_undef;
    }

    NexusMalloc(pnx_gp_new, gp, nexus_global_pointer_t *,
		sizeof(nexus_global_pointer_t));
    nexus_global_pointer(gp, SvRV(perlvar));
    /*
      Think about this some more. Stipulations about what gps can
      point to?
      */
    /*
     * The creation of a global pointer doesn't bump the refcount:
     * it is *just an address*. Refcount-bumps happen with the
     * RemoteReference/RRT mechanism.
     */
    /*
    SvREFCNT_inc(SvRV(perlvar));
    */

    gpref = sv_setref_iv(newSVsv(&sv_undef), clas, (IV) gp);
    
    return gpref;
}

void pnx_gp_DESTROY(SV *self)
{
    nexus_global_pointer_t *gp;

    gp = pnx_gp_get_gp(self, "pnx_gp_DESTROY");

#if defined(PNX_DEBUG) && defined(NEXUS_DEBUG)
    printf("Destroying gp %s\n", nexus_global_pointer_string(gp));
#endif
    nexus_destroy_global_pointer(gp);
    NexusFree(gp);
}

char *pnx_gp_as_string(SV *self)
{
    nexus_global_pointer_t *gp;

    gp = pnx_gp_get_gp(self, "pnx_gp_as_string");

    if (gp == NULL)
	return "";
    else
    {
#ifdef NEXUS_DEBUG
	return (char *) nexus_global_pointer_string(gp);
#else
	static char buf[1024];
	sprintf(buf, "GP(%x)", gp);
	return buf;
#endif
    }
}
void pnx_gp_print(SV *self)
{
    nexus_global_pointer_t *gp;

    gp = pnx_gp_get_gp(self, "pnx_gp_print");

#ifdef NEXUS_DEBUG
    printf("GP=%s\n", nexus_global_pointer_string(gp));
#else
    printf("GP=%x\n", gp);
#endif
}

int pnx_gp_is_local(SV *self)
{
    nexus_global_pointer_t *gp;

    gp = pnx_gp_get_gp(self, "pnx_gp_is_local");

    return nexus_global_pointer_to_current_context(gp);
}

int pnx_gp_test(SV *self)
{
    nexus_global_pointer_t *gp;

    gp = pnx_gp_get_gp(self, "pnx_gp_is_local");

    return nexus_test_global_pointer(gp);
}

SV *pnx_gp_dereference(SV *self)
{
    SV *gpv, *var, *varref;
    nexus_global_pointer_t *gp;

    gp = pnx_gp_get_gp(self, "pnx_gp_dereference");

    var = (SV *) nexus_convert_global_pointer_address(gp);
    varref = newRV(var);
    return varref;
}

nexus_global_pointer_t *pnx_gp_get_gp(SV *self, char *func)
{
    nexus_global_pointer_t *gp;
    SV *bufaddr;
    HV *stash;
    char *package;
    
    if (!SvROK(self))
    {
	croak("Must pass Nexus::GlobalPointer to %s (1)", func);
	return (nexus_global_pointer_t *) NULL;
    }

    bufaddr = SvRV(self);

    if (!SvOK(bufaddr))
    {
	croak("Must pass Nexus::GlobalPointer to %s (2)", func);
	return (nexus_global_pointer_t *) NULL;
    }

    stash = SvSTASH(bufaddr);
    if (!stash)
    {
	croak("Must pass Nexus::GlobalPointer to %s (3)", func);
	return (nexus_global_pointer_t *) NULL;
    }

    package = HvNAME(stash);

    if (strcmp(package, "Nexus::GlobalPointer") != 0)
    {
	croak("Must pass Nexus::GlobalPointer to %s (4)", func);
	return (nexus_global_pointer_t *) NULL;
    }
    gp = (nexus_global_pointer_t *) SvIV(bufaddr);
    return gp;
}
