#include "EXTERN.h"
#include "perl.h"
#include "nexus.h"
#include "pnx.h"

#define PENDING_RPCS "Nexus::PendingRPCList"

static void async_rpc_handler(void *address, nexus_buffer_t *buffer,
				  char *handler_name, int handler_id);

static void rpc_handler(void *address, nexus_buffer_t *buffer);
#define RPC_HANDLER_NAME "Nexus::rpc_handler"
#define RPC_HANDLER_HASH 780

static void rpc_return_handler(void *address, nexus_buffer_t *buffer);
#define RPC_RETURN_HANDLER_NAME "Nexus::rpc_return_handler"
#define RPC_RETURN_HANDLER_HASH 526

static void rrt_incr_handler(void *address, nexus_buffer_t *buffer);
#define	    RRT_INCR_HANDLER_NAME "Nexus::RRT::incr"
#define     RRT_INCR_HANDLER_HASH 418

static void rrt_decr_handler(void *address, nexus_buffer_t *buffer);
#define	    RRT_DECR_HANDLER_NAME "Nexus::RRT::decr"
#define     RRT_DECR_HANDLER_HASH 404

static nexus_handler_t pnx_handlers[] =
{ \
  {"NexusUnknownHandler",
       NEXUS_UNKNOWN_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) async_rpc_handler},
  {RRT_INCR_HANDLER_NAME,
       RRT_INCR_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) rrt_incr_handler},
  {RRT_DECR_HANDLER_NAME,
       RRT_DECR_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) rrt_decr_handler},
  {RPC_HANDLER_NAME,
       RPC_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) rpc_handler},
  {RPC_RETURN_HANDLER_NAME,
       RPC_RETURN_HANDLER_HASH,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) rpc_return_handler},
  {(char *) NULL,
       0,
       NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) NULL},
};

/*
#undef SvREFCNT_dec
#define SvREFCNT_dec(x)
*/

/*
 * NexusBoot()
 */
int NexusBoot()
{
    int rc;
    
    nexus_register_handlers(pnx_handlers);

    pnx_boot();
    pnx_fd_boot();
    pnx_mb_boot();
    pnx_smb_boot();
    pnx_gp_boot();
    pnx_attach_boot();
    pnx_fault_boot();

    do {
	dSP ;
	PUSHMARK(sp) ;
	PUTBACK ;
	perl_call_pv("Nexus::boot", G_DISCARD | G_EVAL);
    } while(0);

    rc = 0;
    
    return (rc);
} /* NexusBoot() */

void pnx_boot()
{
    nexus_register_handlers(pnx_handlers);

}

SV *pnx_create_context(SV *gpref, char *pathname, int size)
{
    croak("Nexus::create_context() not defined in NexusLite");
}

SV *pnx_init_rsr(SV *gp_ref, char *subroutine)
{
    nexus_global_pointer_t *gp;
    pnx_buffer_t *buffer;
    SV *buf_val_ref;

    NexusMalloc(pnx_init_rsr, buffer, pnx_buffer_t *,
		sizeof(pnx_buffer_t));

    strcpy(buffer->subname, subroutine);

    if (sv_isa(gp_ref, "Nexus::GlobalPointer"))
    {
	gp = pnx_gp_get_gp(gp_ref, "pnx_init_rsr");
    }
    else if (sv_isa(gp_ref, "Nexus::RemoteReference"))
    {
	gp_ref = pnx_rrt_gp(gp_ref);
	gp = pnx_gp_get_gp(gp_ref, "pnx_init_rsr");
    }
    else
    {
	croak("Must pass global pointer or remote reference to init_rsr");
    }

#if defined(PNX_DEBUG) && defined(NEXUS_DEBUG)
    printf("init_rsr sub=%s gp=%s\n", subroutine,
	   nexus_global_pointer_string(gp));
#endif

    nexus_init_remote_service_request(&buffer->buffer, gp,
				      buffer->subname,
				      nexus_handler_hash(buffer->subname));
    
    buf_val_ref = sv_setref_iv(newSVsv(&sv_undef),
			       "Nexus::MessageBuffer", (IV) buffer);
    return buf_val_ref;
}

SV *pnx__poll(int block)
{
    if (block)
    {
	nexus_blocking_poll();
    }
    else
    {
	nexus_poll();
    }
}

/*
 * Convert an address that comes in as the target of a remote
 * service request into the local perl ref. There's some serious
 * magic here; I don't remember all the implications. I just remember
 * playing with this a lot.
 *
 * If address is null, return undef.
 *
 * If address is a non-object pointer, return a reference to the data
 * object.
 *
 * If address is a Nexus::RRT, return the ref in the trap.
 *
 * If address is a Nexus::GlobalPointer, dereference the global pointer
 * and return that.
 *
 * Otherwise, just return a ref to the object.
 */
static SV *address_to_ref(SV *address)
{
    SV *addr_ref;
    if (address == NULL)
    {
	addr_ref = newSVsv(&sv_undef);
/*	printf("address_to_ref: got null\n");*/
    }
    else if (!SvOBJECT(address))
    {
	addr_ref = newRV(address);
/*	printf("address_to_ref: got nonobject\n");*/
    }
    else
    {
	char *name;

	name = HvNAME(SvSTASH(address));
	if (strcmp(name, "Nexus::RRT") == 0)
	{
	    /*
	     * The ref is $$self[1] in the RRT. Just yank it out.
	     */
	    addr_ref = *av_fetch((AV *) address, 1, 0);
/*	    printf("address_to_ref: got RRT\n");
	    pnx_sv_dump(address);
	    printf("got ref:\n");
	    pnx_sv_dump(addr_ref); */
	}
	else if (strcmp(name, "Nexus::GlobalPointer") == 0)
	{
	    addr_ref = pnx_gp_dereference(address);
	    addr_ref = sv_2mortal(addr_ref);
/*	    printf("address_to_ref: got gp\n");*/
	}
	else
	{
	    addr_ref = newRV(address);
/*	    printf("address_to_ref: got other object %s\n",name);*/
	}
    }
    return addr_ref;
}

SV *sv_list_to_object(int n_ptrs, SV **ptrs, char *objname)
{
    AV *av;
    HV *stash;
    SV *obj;
    int i;

    stash = gv_stashpv(objname, TRUE);

    av = av_make(n_ptrs, ptrs);

    obj = newRV((SV *) av);
    obj = sv_bless(obj, stash);

    /*
     * Decrement the count in the array, so that the
     * object is the only thing that has a ref to it.
     */
    SvREFCNT_dec(av);

    /*
    printf("created object %s ", objname);
    for (i = 0; i < n_ptrs; i++)
    {
	printf("%x ", ptrs[i]);
    }
    printf(" obj=%x\n", obj);
    */

    return obj;
}

/*
 * Unknown message handler. Async rpc's come in here.
 *
 * We go ahead and do the address->SV conversion: since we're here,
 * we must have been invoked on a perl data object somewhere.
 *
 * We create the array [name, ref, buffer] and bless it into
 * Nexus::AsyncHandler. Ref to the array is pushed onto
 * @Nexus::PendingRPCList.
 */
static void async_rpc_handler(void *address, nexus_buffer_t *buffer,
				  char *handler_name, int handler_id)
{
    SV *obj;
    AV *rpcs;
    SV *ptrs[3];
    nexus_stashed_buffer_t *stashed;

    NexusMalloc(pnx_async_rpc_handler, stashed, nexus_stashed_buffer_t *,
		sizeof(nexus_stashed_buffer_t));
    
    nexus_stash_buffer(buffer, stashed);


    /*
    printf("async_rpc_handler: address=%x buffer=%x stashed=%x name=%s\n",
	   address, buffer, stashed, handler_name);
    */

    ptrs[0] = address_to_ref((SV *) address);
    ptrs[1] = newSVpv(handler_name, strlen(handler_name));
    ptrs[2] = sv_setref_iv(newSVsv(&sv_undef),
			  "Nexus::StashedMessageBuffer",
			  (IV) stashed);

    obj = sv_list_to_object(3, ptrs, "Nexus::AsyncHandler");

    SvREFCNT_dec(ptrs[1]);
    SvREFCNT_dec(ptrs[2]);

    rpcs = perl_get_av(PENDING_RPCS, TRUE);

    av_push(rpcs, obj);

    /*
    {
	printf("WALK_URL: rpcs\n");
	pnx_sv_dump((SV *) rpcs);
	printf("WALK_URL: obj\n");
	pnx_sv_dump(obj);
	pnx_sv_dump(ptrs[0]);
	pnx_sv_dump(ptrs[1]);
	pnx_sv_dump(ptrs[2]);
    }
    */
}

/*
 * RPC handler.
 */
static void rpc_handler(void *address, nexus_buffer_t *buffer)
{
    SV *obj;
    AV *rpcs;
    SV *ptrs[2];
    nexus_stashed_buffer_t *stashed;

    NexusMalloc(pnx_rpc_handler, stashed, nexus_stashed_buffer_t *,
		sizeof(nexus_stashed_buffer_t));
    
    nexus_stash_buffer(buffer, stashed);

    ptrs[0] = address_to_ref((SV *) address);
    ptrs[1] = sv_setref_iv(newSVsv(&sv_undef),
			  "Nexus::StashedMessageBuffer",
			  (IV) stashed);

/*
    printf("rpc_handler: address=%x buffer=%x stashed=%x\n",
	   address, buffer, stashed);
*/

    obj = sv_list_to_object(2, ptrs, "Nexus::RPCHandler");

    SvREFCNT_dec(ptrs[1]);

    rpcs = perl_get_av(PENDING_RPCS, TRUE);

    av_push(rpcs, obj);
/*
    printf("rpc_handler: obj=\n");
    pnx_sv_dump(obj);
    printf("smb=\n");
    pnx_sv_dump(ptrs[1]);
*/    
}

/*
 * RPC return value handler.
 */
static void rpc_return_handler(void *address, nexus_buffer_t *buffer)
{
    SV *obj;
    AV *rpcs;
    SV *ptrs[2];
    nexus_stashed_buffer_t *stashed;

    NexusMalloc(rpc_return_handler, stashed, nexus_stashed_buffer_t *,
		sizeof(nexus_stashed_buffer_t));
    
    nexus_stash_buffer(buffer, stashed);

    ptrs[0] = address_to_ref((SV *) address);
    ptrs[1] = sv_setref_iv(newSVsv(&sv_undef),
			  "Nexus::StashedMessageBuffer",
			  (IV) stashed);

    obj = sv_list_to_object(2, ptrs, "Nexus::RPCReturnHandler");

    SvREFCNT_dec(ptrs[1]);

    rpcs = perl_get_av(PENDING_RPCS, TRUE);

    av_push(rpcs, obj);
/*
    printf("rpc_return_handler: obj=\n");
    pnx_sv_dump(obj);
    printf("smb=\n");
    pnx_sv_dump(ptrs[1]);
*/
}

/*
 * Increment an RRT refcount.
 *
 * We *could* just do it here, but we'll create a Nexus::RRTIncr object
 * to encapsulate this anyway.
 */
static void rrt_incr_handler(void *address, nexus_buffer_t *buffer)
{
    SV *obj;
    AV *rpcs;
    SV *ptrs[2];
    int value;

    nexus_get_int(buffer, &value, 1);

    /*
     * Make addr a ref to the RRT itself
     */
    ptrs[0] = newRV(address);
    ptrs[1] = newSViv(value);

    obj = sv_list_to_object(2, ptrs, "Nexus::RRTIncr");

    SvREFCNT_dec(ptrs[0]);
    SvREFCNT_dec(ptrs[1]);
    
    rpcs = perl_get_av(PENDING_RPCS, TRUE);

    av_push(rpcs, obj);

    SvREFCNT_inc(obj);

/*
    printf("incr_handler: obj=\n");
    pnx_sv_dump(obj);
    printf("addr=\n");
    pnx_sv_dump(ptrs[0]);
    printf("value=\n");
    pnx_sv_dump(ptrs[1]);
    printf("rpcs=\n");
    pnx_sv_dump(rpcs);
*/
}

static void rrt_decr_handler(void *address, nexus_buffer_t *buffer)
{
    SV *obj;
    AV *rpcs;
    SV *ptrs[2];
    int value;

    nexus_get_int(buffer, &value, 1);

    /*
     * Make addr a ref to the RRT itself
     */
    ptrs[0] = newRV(address);
    ptrs[1] = newSViv(value);

    obj = sv_list_to_object(2, ptrs, "Nexus::RRTDecr");

    SvREFCNT_dec(ptrs[0]);
    SvREFCNT_dec(ptrs[1]);

    rpcs = perl_get_av(PENDING_RPCS, TRUE);

    av_push(rpcs, obj);
}
