#include "EXTERN.h"
#include "perl.h"
#include "nexus.h"
#include "pnx.h"

#undef PNX_DEBUG 

#ifndef DBL_DIG
#define DBL_DIG 15
#endif

SV *pnx_rref_new(SV *obj, char *type)
{
    SV *rref;
    SV *cls_sv, *type_sv;
    char *cls;
    STRLEN len;
    dSP;
    /* Need the following for XPUSHp. */
    /* dATARGET; */
    int count;

    cls = "Nexus::RemoteReference";
    len = strlen(cls);

    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(sv_2mortal(newSVpv(cls, len)));
    XPUSHs(obj);
    XPUSHs(sv_2mortal(newSVpv(type, strlen(type))));
    PUTBACK;
    count = perl_call_method("new", G_SCALAR | G_EVAL);
    SPAGAIN;
    if (count != 1)
	croak("Big trouble");
    rref = POPs;
    SvREFCNT_inc(rref);
    PUTBACK;
    FREETMPS;
    LEAVE;
    return rref;
}

SV *pnx_rrt_new(SV *obj)
{
    SV *rrt;
    SV *cls_sv;
    char *cls;
    STRLEN len;
    dSP;
    /* Need the following for XPUSHp. */
    /* dATARGET; */
    int count;

    cls = "Nexus::RRT";
    len = strlen(cls);

    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(sv_2mortal(newSVpv(cls, len)));
    XPUSHs(obj);
    PUTBACK;
    count = perl_call_method("new", G_SCALAR | G_EVAL);
    SPAGAIN;
    if (count != 1)
	croak("Big trouble");
    rrt = POPs;
    SvREFCNT_inc(rrt);
    PUTBACK;
    FREETMPS;
    LEAVE;
    return rrt;
}

SV *pnx_rrt_gp(SV *rrt)
{
    SV *gp;
    
    dSP;
    int count;
    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(rrt);
    PUTBACK;
    count = perl_call_method("gp", G_SCALAR | G_EVAL);
    SPAGAIN;
    if (count != 1)
	croak("Big trouble");
    gp = POPs;
    SvREFCNT_inc(gp);
    PUTBACK;
    FREETMPS;
    LEAVE;
    return gp;
}

char *pnx_rrt_type(SV *rrt)
{
    char *type;
    static char buf[2048];
    
    dSP;
    int count;
    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(rrt);
    PUTBACK;
    count = perl_call_method("type", G_SCALAR | G_EVAL);
    SPAGAIN;
    if (count != 1)
	croak("Big trouble");
    type = POPp;
    if (strlen(type) > (sizeof(buf) - 1))
	croak("rrt_type: name %s too long", type);
    strcpy(buf, type);
    PUTBACK;
    FREETMPS;
    LEAVE;
    return buf;
}

SV *pnx_rrt_ref(SV *rrt)
{
    SV *ref;
    
    dSP;
    int count;
    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(rrt);
    PUTBACK;
    count = perl_call_method("ref", G_SCALAR | G_EVAL);
    SPAGAIN;
    if (count != 1)
	croak("Big trouble");
    ref = POPs;
    SvREFCNT_inc(ref);
    PUTBACK;
    FREETMPS;
    LEAVE;
    return ref;
}

void pnx_rrt_incr(SV *rrt, int count)
{
    dSP;
    SV *countsv;

    countsv = newSViv(count);
    
    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(rrt);
    XPUSHs(sv_2mortal(countsv));
    PUTBACK;
    perl_call_method("incr", G_SCALAR | G_EVAL);
    SPAGAIN;
    PUTBACK;
    FREETMPS;
    LEAVE;
}

void pnx_rrt_decr(SV *rrt, int count)
{
    dSP;
    SV *countsv;

    countsv = newSViv(count);
    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(rrt);
    XPUSHs(sv_2mortal(countsv));
    PUTBACK;
    perl_call_method("decr", G_SCALAR | G_EVAL);
    SPAGAIN;
    PUTBACK;
    FREETMPS;
    LEAVE;
}

