#include "EXTERN.h"
#include "perl.h"
#include "nexus.h"
#include "pnx.h"

nexus_stashed_buffer_t *_pnx_smb_get_buffer(SV *self, char *func);

void pnx_smb_boot(void)
{
}

void pnx_smb_get_int(SV *self, SV *intref)
{
    nexus_stashed_buffer_t *buffer;
    int ival;
    SV *intvar;

    buffer = _pnx_smb_get_buffer(self, "pnx_smb_get_int");

    nexus_get_stashed_int(buffer, &ival, 1);

    if (!SvROK(intref))
    {
	croak("Must pass reference to pnx_smb_get_int");
	return;
    }

    intvar = SvRV(intref);
    sv_setiv(intvar, ival);
}

void pnx_smb_get_double(SV *self, SV *doubleref)
{
    nexus_stashed_buffer_t *buffer;
    double ival;
    SV *doublevar;

    buffer = _pnx_smb_get_buffer(self, "pnx_smb_get_double");

    nexus_get_stashed_double(buffer, &ival, 1);

    if (!SvROK(doubleref))
    {
	croak("Must pass reference to pnx_smb_get_double");
	return;
    }

    doublevar = SvRV(doubleref);
    sv_setnv(doublevar, ival);
}

void pnx_smb_get_string(SV *self, SV *stringref, int len)
{
    nexus_stashed_buffer_t *buffer;
    char *sval;
    SV *stringvar;

    buffer = _pnx_smb_get_buffer(self, "pnx_smb_get_string");

    NexusMalloc(pnx_smb_get_string, sval, char *, len);

    nexus_get_stashed_char(buffer, sval, len);

    if (!SvROK(stringref))
    {
	croak("Must pass reference to pnx_smb_get_string");
	return;
    }

    stringvar = SvRV(stringref);
    sv_setpvn(stringvar, sval, len);
    NexusFree(sval);
}

void pnx_smb_get_global_pointer(SV *self, SV *gpref)
{
    nexus_stashed_buffer_t *buffer;
    nexus_global_pointer_t *gp;
    SV *gpv, *gpvr, *gpvar, *gpvrb;
    HV *stash;

    buffer = _pnx_smb_get_buffer(self, "pnx_smb_get_global_pointer");

    NexusMalloc(pnx_smb_get_global_pointer, gp, nexus_global_pointer_t *,
		sizeof(nexus_global_pointer_t));
    nexus_get_stashed_global_pointer(buffer, gp, 1);

    if (!SvROK(gpref))
    {
	croak("Must pass reference to pnx_smb_get_global_pointer");
	return;
    }

/*    printf("created global pointer %s\n", nexus_global_pointer_string(gp)); */

    gpvar = SvRV(gpref);
    sv_setref_iv(gpvar, "Nexus::GlobalPointer", (long) gp);
}

void pnx_smb_DESTROY(SV *self)
{
    nexus_stashed_buffer_t *buffer;

    buffer = _pnx_smb_get_buffer(self, "pnx_smb_DESTROY");

/*    printf("free smb %x\n", buffer);*/
    nexus_free_stashed_buffer(buffer);
    NexusFree(buffer);
}

nexus_stashed_buffer_t *_pnx_smb_get_buffer(SV *self, char *func)
{
    nexus_stashed_buffer_t *buffer;
    SV *bufaddr;
    HV *stash;
    char *package;
    
    if (!sv_isa(self, "Nexus::StashedMessageBuffer"))
    {
	croak("Must pass Nexus::StashedMessageBuffer to %s", func);
	return NULL;
    }

    bufaddr = SvRV(self);

    buffer = (nexus_stashed_buffer_t *) SvIV(bufaddr);
    return buffer;
}
